"""The image subcommand group."""

import asyncio

import typer
from rich import print
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from smpclient import SMPClient
from smpclient.generics import error, success
from smpclient.requests.image_management import ImageStatesRead
from smpclient.transport.serial import SMPSerialTransport
from typing_extensions import Annotated

from smpmgr import const
from smpmgr.common import connect_with_spinner

app = typer.Typer(name="image", help="The SMP Image Management Group.")


@app.command()
def state_read(address: const.Address) -> None:
    """Request to read the state of FW images on the SMP Server."""

    smpclient = SMPClient(SMPSerialTransport(), address)

    async def f() -> None:
        if await connect_with_spinner(smpclient) is False:
            print("Timeout")
            return

        with Progress(
            SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True
        ) as progress:
            progress.add_task(description="Requested image states...", total=None)

            try:
                r = await asyncio.wait_for(
                    smpclient.request(ImageStatesRead()), timeout=const.CONNECT_TIMEOUT_S  # type: ignore # noqa
                )
            except asyncio.TimeoutError:
                print("Timeout")
                return

        if error(r):
            print(r)
        elif success(r):
            if len(r.images) == 0:
                print("No images on device!")
            for image in r.images:
                print(image)
            if r.splitStatus is not None:
                print(f"splitStatus: {r.splitStatus}")
        else:
            raise Exception("Unreachable")

    asyncio.run(f())


async def upload_with_progress_bar(smpclient: SMPClient, file: typer.FileBinaryRead) -> None:
    """Animate a progress bar while uploading the FW image."""

    with Progress(
        TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeRemainingColumn(),
    ) as progress:
        image = file.read()
        file.close()
        task = progress.add_task("Uploading", total=len(image), filename=file.name, start=True)
        async for offset in smpclient.upload(image):
            progress.update(task, completed=offset)


@app.command()
def upload(
    address: const.Address,
    file: Annotated[typer.FileBinaryRead, typer.Argument(help="Path to FW image")],
) -> None:
    """Upload a FW image."""
    smpclient = SMPClient(SMPSerialTransport(), address)

    async def f() -> None:
        if await connect_with_spinner(smpclient) is False:
            print("Timeout")
            return

        await upload_with_progress_bar(smpclient, file)

    asyncio.run(f())



"""Entry point for the `smpmgr` application."""

import asyncio

import typer
from rich import print
from smp.os_management import OS_MGMT_RET_RC
from smpclient import SMPClient
from smpclient.generics import error, success
from smpclient.requests.os_management import ResetWrite
from smpclient.transport.serial import SMPSerialTransport
from typing_extensions import Annotated

from smpmgr import const, image_management, os_management
from smpmgr.common import connect_with_spinner
from smpmgr.image_management import upload_with_progress_bar

app = typer.Typer()
app.add_typer(os_management.app)
app.add_typer(image_management.app)


@app.command()
def upgrade(
    address: const.Address,
    file: Annotated[typer.FileBinaryRead, typer.Argument(help="Path to FW image")],
) -> None:
    """Upload a FW image, mark it for next boot, and reset the device."""
    smpclient = SMPClient(SMPSerialTransport(), address)

    async def f() -> None:
        if await connect_with_spinner(smpclient) is False:
            print("Timeout")
            return

        await upload_with_progress_bar(smpclient, file)

        r = await smpclient.request(ResetWrite())  # type: ignore
        if error(r):
            if r.rc != OS_MGMT_RET_RC.OK:
                print(r)
                return
        elif success(r):
            pass
        else:
            raise Exception("Unreachable")

        print("Upgrade complete.  The device may take a few minutes to complete FW swap.")

    asyncio.run(f())



"""Common CLI helpers from rich, typer, click, etc."""
import asyncio

from rich.progress import Progress, SpinnerColumn, TextColumn
from smpclient import SMPClient

from smpmgr import const


async def connect_with_spinner(smpclient: SMPClient) -> bool:
    """Animate a spinner while waiting to connect to the SMP Server."""
    with Progress(
        SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True
    ) as progress:
        progress_connect = progress.add_task(
            description=f"Connecting to {smpclient._address}...", total=None
        )
        try:
            await asyncio.wait_for(smpclient.connect(), timeout=const.CONNECT_TIMEOUT_S)
            progress.remove_task(progress_connect)
            return True
        except asyncio.TimeoutError:
            return False



"""Constants for `smpmgr`."""
from typing import Final

import typer
from typing_extensions import Annotated

CONNECT_TIMEOUT_S: Final = 5.0

Address = Annotated[str, typer.Option(help="Serial port, e.g /dev/ttyACM0, COM1, etc.")]



import asyncio

import typer
from rich import print
from smpclient import SMPClient
from smpclient.requests.os_management import EchoWrite, ResetWrite
from smpclient.transport.serial import SMPSerialTransport

from smpmgr import const

app = typer.Typer(name="os", help="The SMP OS Management Group.")


@app.command()
def echo(address: str, message: str) -> None:
    """Request that the SMP Server echo the given message."""
    s = SMPClient(SMPSerialTransport(), address)

    async def f() -> None:
        await s.connect()
        r = await s.request(EchoWrite(d=message))  # type: ignore
        print(r)

    asyncio.run(f())


@app.command()
def reset(address: const.Address) -> None:
    """Request that the SMP Server reset the device."""
    s = SMPClient(SMPSerialTransport(), address)

    async def f() -> None:
        await s.connect()
        r = await s.request(ResetWrite())  # type: ignore
        print(r)

    asyncio.run(f())
